//
//    File: Slice2.hh
//
//    Fast slice checker
//
#ifndef SLICE2_HH
#define SLICE2_HH

#include "DMTBase.hh"
#include "TrigClient.hh"
#include "Interval.hh"
#include "Complex.hh"
#include <list>
#include <string>

class ChanSlice;
class Time;

/**  @name Slice2
  *  Slice2 searches specified channels for repeated fixed-length 
  *  data sequences. This detects DAQ errors resulting in a DAQ slice
  *  not being replaced by new data. 
  *
  *  Repeated slices are identified by comparing each data slice to all 
  *  previous slices up to a specified maximum offset. By default the 
  *  slice length is 62.5ms and the maximum offset is 250ms. These 
  *  parameters may be respecified on a channel-by-channel basis within
  *  the configuration file. A slice error is flagged if more than a 
  *  specified minimum number of channels is found to have a repeated 
  *  slice. The minimum channel count is set with the #nskip# argument
  *  on the command line and defaults to 2.
  *
  *  The following trigger types are generated by Slice2:
  *
  *  \begin{center}
  *  \begin{tabular}{llll}
  *    Name & Sub-type & mask & Meaning \\
  *    FrameTimeError & TimeIsZero & 0x40 & Frame time-stamp is zero. \\
  *    FrameTimeError & MisAligned & 0x20 & Frame time-stamp is not on a 
  *                                         1-second boundary. \\
  *    FrameTimeError & Not_1_second & 0x10 & Frame is not 1 second long. \\
  *    ChannelReadOutError & <channel> & 0x08 & Error flagged in FrAdcData.\\
  *    TimeSkewError & <channel> & 0x04 & Invalid Adc time offset #(0<t<1s)#\\
  *    TimeIntervalError & <channel> & 0x02 & Data series duration #># frame 
  *                                           length. \\
  *    TimeSliceError & MultiChannel & 0x01 & Candidate misplaced time slice.
  *  \end{tabular}
  *  \end{center}
  *
  *  \begin{center}{\bf {\Large The Slice2 Configuration File}}\end{center}
  *
  *  The Slice2 configuration file specifies the channels to be monitored 
  *  and parameters affecting which slices are compared and the generation 
  *  of triggers for each channel. Each line of the configuration file 
  *  specifies the parameters for a single channel and contains the 
  *  following fields:
  *
  *  <channel-name> <slice-time> <max-offset>
  *
  *  The meaning of each field is as follows:
  *
  *  \begin{center}
  *  \begin{tabular}{ll}
  *    <channel-name> & Name of the channel to be monitored. \\
  *    <slice-time> & Time slice length [0.0625]. \\
  *    <max-offset> & Maximum offset time [0.25].
  *  \end{tabular}
  *  \end{center}
  *
  *
  *  \begin{center}{\bf {\Large Running Slice2}}\end{center}
  *
  *  The syntax of the Slice2 command is as follows:
  *
  *  \begin{center}
  *  \begin{verbatim}
Slice2 [-partition <pname>] [-infile <file>] [cfile <config>] \
       [+trig[:<mask>]] [nskip <skip-min>] [-debug <dbg-level>] 
     \end{verbatim}
  *  \end{center}
  *
  *  Where the arguments have the following meaning:
  *
  *  \begin{center}
  *  \begin{tabular}{ll}
  *    <pname>        & Shared memory partition name with data to be read \\
  *    <file>         & Input frame file (exclusive of #<part-name>#) \\
  *    <dbg-level>    & Debug level \\
  *    <config>       & Configuration file name. \\
  *    +trig[:<mask>] & enables trigger generation with mask #<mask>#. \\
  *    <skip-min>     & Minimum channels to define a slice error [2].\\
  *  \end{tabular}
  *  \end{center}
  *
  *  The partition name is mutually exclusive with the input frame file 
  *  name. If both are specified, data are read from the specified shared 
  *  memory partition. The debug level defaults to 0 (no debug messages). 
  *  Any other value for #<dbg-level># will cause debugging messages to be 
  *  printed to cout/cerr.
  *
  *  {\bf Modifying the online configuration} \\
  *  The Slice2 configuration file can be modified while Slice2 is running.
  *  This is accomplished by editing the current Slice2 configuration file,
  *  usually {\bf ~ops/pars/Slice2.conf}. Once the file has been modified, 
  *  Slice2 can be made to read in the new configuration by signaling the
  *  running process with SIGUSR1. When the signal is caught by the process,
  *  Slice2 will write out the status files with whatever statistics have 
  *  already been collected, read the configuration file and restart 
  *  processing. The SIGUSR1 signal is delivered with the following command:
  *
  *  \begin{center}#pkill -USR1 Slice2#\end{center}
  *
  *  \begin{center}{\bf {\Large Slice2 Reports and Other Output}}\end{center}
  *
  *  {\bf Status Reports} \\
  *
  *  Slice2 produces a status report every 5 minutes. The report is written
  *  in the html output directory if #DMTHTMLOUT# is defined, otherwise in 
  *  the current working directory. The report file name is Slice2.report.
  *  The report file contains a list of global statistics followed by a 
  *  section containing statistics for each channel. The global statistics
  *  contain the following information:
  *
  *  \begin{tabular}{|l|l|}
  *     Slice time & Time slice used by this program \\
  *     Maximum time shift & Maximum time shift detected by this monitor. \\
  *     Total frames read & Total number of frames read \\
  *     Total triggers & Number of triggers generated since the last 
  *        (re)start of Slice2. \\
  *     Total rate & Trigger rate since the last (re)start of Slice2 \\
  *     Triggers in last 300s & Number of triggers generated in the last 
  *        300s \\
  *     Rate in last 300s & Generated trigger rate in the last 300s. \\
  *     Trigger multiplicity & Number of channels that must have a slice 
  *        error for a trigger to be generated \\
  *     Sum singles rate & Sum of singles rates in all channels \\
  *     Uncorrelated doubles & Expected number of double coincidences 
  *        (random triggers) assuming that the channel slice errors are 
  *        uncorrelated.
  *  \end{tabular}
  *
  *  @memo Look for duplicated data
  *  @author J. Zweizig
  *  @version 3.1; Modified April 5, 2006
  */
//{
//}

//======================================  Monitor class definition
class Slice2 : public DMTBase, TrigClient {
  public:
    Slice2(int argc, const char *argv[]);
    ~Slice2();
    void ProcessFrame(const FrameCPP::FrameH* frame);
    void Attention(void);
    bool ReadConfig(void);
    void Report(void);
    void Reset(void);
    bool trigEnable(int mask) const;

  private:
    typedef std::list<ChanSlice> ChanList;
    typedef ChanList::iterator channel_iter;

  private:
    std::string mConfig;
    bool     mGenTrig;
    bool     mAlign;
    int      mSkipMin;
    int      mTrigMask;
    Interval mGroupSecs;

    ChanList mChannel;
    int      mFrameCt;
    Interval mSumTimes;
    Interval mLastTime;
    int      mTrigCt;
    int      mLastTrig;
};

//======================================  Slice2 inline methods.
inline bool 
Slice2::trigEnable(int mask) const {
    return mGenTrig && (mTrigMask & mask) != 0;
}

//
//    ChanSlice Class
//
//    The ChanSlice class carries the status of each channel to be followed
//    by the slicer.
//
class ChanSlice {
  typedef unsigned int index_t;
  typedef unsigned int size_t;
  public:
    //----------------------------------  Constructors, Destructors.
    ChanSlice(const char *name);
    ChanSlice(const ChanSlice& x);
    ~ChanSlice();
    ChanSlice& operator=(const ChanSlice& x);

    //----------------------------------  Scan the data
    /*   Scan the data for a shifted data slice.
      *  @param nw   number of data words.
      *  @param data data array to be scanned.
      *  @param t0   data start time.
      *  @return Number of overlapping slices found.
      */
    int Scan(size_t nw, const short *data);
    int Scan(size_t nw, const float *data);
    int Scan(size_t nw, const fComplex *data);

    //----------------------------------  Accessors
    const char* getChannel() const;
    double  getErrorRate(void) const;
    index_t getErrorTot(void) const;
    index_t getNTest(void) const;
    double  getSingleRate(void) const;
    double  getTriggerRate(void) const;

    //----------------------------------  Mutators
    void countTrigger(void);
    void incError(void);
    void resetStats(void);
    void saveData(const short* p, size_t N);
    void saveData(const float* p, size_t N);
    void saveData(const fComplex* p, size_t N);
    void freeData(void);
    void setDebug(index_t level);
    void setSampleRate(double Rate);
    void setSliceTime(Interval Slice, Interval maxOff);

  public:
    class stats {
    public:
        stats(void);
        void clear(void);
        stats& operator+=(const stats& x);
        void  inc_test(void);
        void  inc_error(void);
        void  inc_sngl(void);
        void  inc_trig(void);
        index_t Test(void) const;
        index_t Error(void) const;
        index_t Single(void) const;
        index_t Trigger(void) const;
    private:
        index_t     mNTest;
        index_t     mNError;
        index_t     mNSingle;
        index_t     mNTrigger;
    };

  private:
    std::string mChanName;
    size_t      mSaved;
    enum {kNone, kShort, kFloat, kCmplx} mSavType;
    void*       mSavData;
    double      mRate;
    Interval    mSliceTime;
    size_t      mNWSlice;  // Number of words in a slice
    size_t      mNCkSlice; // Number of slices to check
    index_t     mDebug;

    //----------------------------------  Statistics
    index_t  mNOverlap;
    stats    mStats;
    stats    mTotal;
};

//======================================  Inline ChanSlice Methods
void 
ChanSlice::stats::clear(void) {
    mNTest    = 0;
    mNError   = 0;
    mNSingle  = 0;
    mNTrigger = 0;
}

inline 
ChanSlice::stats::stats(void) {
    clear();
}

inline ChanSlice::stats& 
ChanSlice::stats::operator+=(const stats& x) {
    mNTest    += x.mNTest;
    mNError   += x.mNError;
    mNSingle  += x.mNSingle;
    mNTrigger += x.mNTrigger;
    return *this;
}

inline void 
ChanSlice::stats::inc_test(void) {
    mNTest++;
}

inline void 
ChanSlice::stats::inc_error(void) {
    mNError++;
}

inline void 
ChanSlice::stats::inc_sngl(void) {
    mNSingle++;
}

inline void 
ChanSlice::stats::inc_trig(void) {
    mNTrigger++;
}

inline ChanSlice::index_t 
ChanSlice::stats::Test(void) const {
    return mNTest;
}

inline ChanSlice::index_t 
ChanSlice::stats::Error(void) const {
    return mNError;
}

inline ChanSlice::index_t 
ChanSlice::stats::Single(void) const {
    return mNSingle;
}
;
inline ChanSlice::index_t 
ChanSlice::stats::Trigger(void) const {
    return mNTrigger;
}

//======================================  Inline ChanSlice Methods
inline void
ChanSlice::countTrigger(void) {
    if (mNOverlap) mStats.inc_trig();
}

inline const char* 
ChanSlice::getChannel(void) const {
    return mChanName.c_str();
}

inline double
ChanSlice::getErrorRate(void) const {
    return mStats.Error()/double(mStats.Test()+mStats.Error());
}

inline ChanSlice::index_t 
ChanSlice::getErrorTot(void) const {
    return mTotal.Error() + mStats.Error();
}

inline ChanSlice::index_t 
ChanSlice::getNTest(void) const {
    return mStats.Test();
}

inline double  
ChanSlice::getSingleRate(void) const {
    if (!mStats.Test()) return 0.0;
    return mStats.Single()/double(mStats.Test());
}

inline double  
ChanSlice::getTriggerRate(void) const {
    if (!mStats.Test()) return 0.0;
    return mStats.Trigger()/double(mStats.Test());
}

inline void
ChanSlice::incError(void) {
    mStats.inc_error();
}

inline void
ChanSlice::setDebug(index_t level) {
    mDebug = level;
}

#ifndef __CINT__

#endif // __CINT__

#endif // SLICE2_HH

